@@ -107,7 +107,7 @@ open_wrapper(SuilHost*      host,
 	dlerror();
 	void* lib = dlopen(path, RTLD_NOW);
 	if (!lib) {
-		SUIL_ERRORF("Unable to open wrap module %s\n", path);
+		SUIL_ERRORF("Unable to open wrap module %s (%s)\n", path, dlerror());
 		return NULL;
 	}
 
@@ -324,5 +324,8 @@ const void*
 suil_instance_extension_data(SuilInstance* instance,
                              const char*   uri)
 {
-	return instance->descriptor->extension_data(uri);
+	if (instance->descriptor->extension_data) {
+		return instance->descriptor->extension_data(uri);
+	}
+	return NULL;
 }
