@@ -21,6 +21,11 @@
 
 #include "./suil_internal.h"
 
+#ifdef HAVE_LV2_1_6_0
+#    include "lv2/lv2plug.in/ns/ext/options/options.h"
+#    include "lv2/lv2plug.in/ns/ext/urid/urid.h"
+#endif
+
 #define SUIL_TYPE_X11_WRAPPER (suil_x11_wrapper_get_type())
 #define SUIL_X11_WRAPPER(obj) (G_TYPE_CHECK_INSTANCE_CAST((obj), SUIL_TYPE_X11_WRAPPER, SuilX11Wrapper))
 
@@ -28,10 +33,15 @@ typedef struct _SuilX11Wrapper      SuilX11Wrapper;
 typedef struct _SuilX11WrapperClass SuilX11WrapperClass;
 
 struct _SuilX11Wrapper {
-	GtkSocket     socket;
-	GtkPlug*      plug;
-	SuilWrapper*  wrapper;
-	SuilInstance* instance;
+	GtkSocket                   socket;
+	GtkPlug*                    plug;
+	SuilWrapper*                wrapper;
+	SuilInstance*               instance;
+#ifdef HAVE_LV2_1_6_0
+	const LV2UI_Idle_Interface* idle_iface;
+	guint                       idle_id;
+	guint                       idle_ms;
+#endif
 };
 
 struct _SuilX11WrapperClass {
@@ -42,11 +52,45 @@ GType suil_x11_wrapper_get_type(void);  // Accessor for SUIL_TYPE_X11_WRAPPER
 
 G_DEFINE_TYPE(SuilX11Wrapper, suil_x11_wrapper, GTK_TYPE_SOCKET)
 
+/**
+   Check if 'swallowed' subwindow is known to the X server.
+
+   Gdk/GTK can mark the window as realized, mapped and visible even though
+   there is no window-ID on the X server for it yet.  Then,
+   suil_x11_on_size_allocate() will cause a "BadWinow" X error.
+*/
+static bool
+x_window_is_valid(SuilX11Wrapper* socket)
+{
+	GdkWindow* window     = gtk_widget_get_window(GTK_WIDGET(socket->plug));
+	Window     root       = 0;
+	Window     parent     = 0;
+	Window*    children   = NULL;
+	unsigned   childcount = 0;
+
+	XQueryTree(GDK_WINDOW_XDISPLAY(window),
+	           GDK_WINDOW_XID(window),
+	           &root, &parent, &children, &childcount);
+	for (unsigned i = 0; i < childcount; ++i) {
+		if (children[i] == (Window)socket->instance->ui_widget) {
+			return true;
+		}
+	}
+	return false;
+}
+
 static gboolean
 on_plug_removed(GtkSocket* sock, gpointer data)
 {
 	SuilX11Wrapper* const self = SUIL_X11_WRAPPER(sock);
 
+#ifdef HAVE_LV2_1_6_0
+	if (self->idle_id) {
+		g_source_remove(self->idle_id);
+		self->idle_id = 0;
+	}
+#endif
+
 	if (self->instance->handle) {
 		self->instance->descriptor->cleanup(self->instance->handle);
 		self->instance->handle = NULL;
@@ -100,7 +144,7 @@ forward_key_event(SuilX11Wrapper* socket,
                   GdkEvent*       gdk_event)
 {
 	GdkWindow* window = gtk_widget_get_window(GTK_WIDGET(socket->plug));
-	GdkScreen* screen = gdk_window_get_screen(window);
+	GdkScreen* screen = gdk_visual_get_screen(gdk_window_get_visual(window));
 
 	XKeyEvent xev;
 	memset(&xev, 0, sizeof(xev));
@@ -112,11 +156,23 @@ forward_key_event(SuilX11Wrapper* socket,
 	xev.state     = gdk_event->key.state;
 	xev.keycode   = gdk_event->key.hardware_keycode;
 
-	XSendEvent (GDK_WINDOW_XDISPLAY(window),
-	            (Window)socket->instance->ui_widget,
-	            False,
-	            NoEventMask,
-	            (XEvent*)&xev);
+	XSendEvent(GDK_WINDOW_XDISPLAY(window),
+	           (Window)socket->instance->ui_widget,
+	           False,
+	           NoEventMask,
+	           (XEvent*)&xev);
+}
+
+static void
+forward_size_request(SuilX11Wrapper* socket,
+                     GtkAllocation*  allocation)
+{
+	GdkWindow* window = gtk_widget_get_window(GTK_WIDGET(socket->plug));
+	if (x_window_is_valid(socket)) {
+		XResizeWindow(GDK_WINDOW_XDISPLAY(window),
+		              (Window) socket->instance->ui_widget,
+		              allocation->width, allocation->height);
+	}
 }
 
 static gboolean
@@ -133,6 +189,20 @@ suil_x11_wrapper_key_event(GtkWidget*   widget,
 	return FALSE;
 }
 
+static void
+suil_x11_on_size_allocate(GtkWidget*     widget,
+                          GtkAllocation* a)
+{
+	SuilX11Wrapper* const self = SUIL_X11_WRAPPER(widget);
+
+	if (self->plug
+	    && GTK_WIDGET_REALIZED(widget)
+	    && GTK_WIDGET_MAPPED(widget)
+	    && GTK_WIDGET_VISIBLE(widget)) {
+		forward_size_request(self, a);
+	}
+}
+
 static void
 suil_x11_wrapper_class_init(SuilX11WrapperClass* klass)
 {
@@ -148,9 +218,13 @@ suil_x11_wrapper_class_init(SuilX11WrapperClass* klass)
 static void
 suil_x11_wrapper_init(SuilX11Wrapper* self)
 {
-	self->plug     = GTK_PLUG(gtk_plug_new(0));
-	self->wrapper  = NULL;
-	self->instance = NULL;
+	self->plug       = GTK_PLUG(gtk_plug_new(0));
+	self->wrapper    = NULL;
+	self->instance   = NULL;
+#ifdef HAVE_LV2_1_6_0
+	self->idle_iface = NULL;
+	self->idle_ms    = 1000 / 30;  // 30 Hz default
+#endif
 }
 
 static int
@@ -160,6 +234,18 @@ wrapper_resize(LV2UI_Feature_Handle handle, int width, int height)
 	return 0;
 }
 
+#ifdef HAVE_LV2_1_6_0
+static gboolean
+suil_x11_wrapper_idle(void* data)
+{
+	SuilX11Wrapper* const wrap = SUIL_X11_WRAPPER(data);
+
+	wrap->idle_iface->idle(wrap->instance->handle);
+	
+	return TRUE;  // Continue calling
+}
+#endif
+
 static int
 wrapper_wrap(SuilWrapper*  wrapper,
              SuilInstance* instance)
@@ -170,11 +256,26 @@ wrapper_wrap(SuilWrapper*  wrapper,
 	wrap->wrapper         = wrapper;
 	wrap->instance        = instance;
 
+#ifdef HAVE_LV2_1_6_0
+	const LV2UI_Idle_Interface* idle_iface = suil_instance_extension_data(
+		instance, LV2_UI__idleInterface);
+	if (idle_iface) {
+		wrap->idle_iface = idle_iface;
+		wrap->idle_id    = g_timeout_add(
+			wrap->idle_ms, suil_x11_wrapper_idle, wrap);
+	}
+#endif
+
 	g_signal_connect(G_OBJECT(wrap),
 	                 "plug-removed",
 	                 G_CALLBACK(on_plug_removed),
 	                 NULL);
 
+	g_signal_connect(G_OBJECT(wrap),
+	                 "size-allocate",
+	                 G_CALLBACK(suil_x11_on_size_allocate),
+	                 NULL);
+
 	return 0;
 }
 
@@ -215,5 +316,31 @@ suil_wrapper_new(SuilHost*      host,
 	suil_add_feature(features, &n_features, LV2_UI__resize,
 	                 &wrapper->resize);
 
+#ifdef HAVE_LV2_1_6_0
+	suil_add_feature(features, &n_features, LV2_UI__idleInterface, NULL);
+
+	// Scan for URID map and options
+	LV2_URID_Map*       map     = NULL;
+	LV2_Options_Option* options = NULL;
+	for (LV2_Feature** f = *features; *f && (!map || !options); ++f) {
+		if (!strcmp((*f)->URI, LV2_OPTIONS__options)) {
+			options = (*f)->data;
+		} else if (!strcmp((*f)->URI, LV2_URID__map)) {
+			map = (*f)->data;
+		}
+	}
+
+	if (map && options) {
+		// Set UI update rate if given
+		LV2_URID ui_updateRate = map->map(map->handle, LV2_UI__updateRate);
+		for (LV2_Options_Option* o = options; o->key; ++o) {
+			if (o->key == ui_updateRate) {
+				wrap->idle_ms = 1000.0f / *(float*)o->value;
+				break;
+			}
+		}
+	}
+#endif
+
 	return wrapper;
 }
